/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  gLite Security Delegation Clients - Main function for the tools
 *
 *  Authors: Zoltan Farkas <Zoltan.Farkas@cern.ch>
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glite/data/delegation/c/delegation-simple.h>

#include <glib.h>
#include <stdio.h>
#include <getopt.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "delegation-main.h"

/**********************************************************************
 * Global variables
 */

/* Location of the transfer service */
static char *service_location;

/* Command line options common to all the tools */
static const char *common_options = "hqs:Vv";

/* If the tool shall keep silence. */
int verbose_flag = 0;
int quiet_flag = 0;

/* Description of the common options */
static const char *common_help =
	"\t-h\t\tPrint this help text and exit.\n"
	"\t-q\t\tQuiet operation.\n"
	"\t-v\t\tVerbose operation.\n"
	"\t-s URL\t\tUse the delegation service at the specified URL.\n"
	"\t-V\t\tPrint the version number and exit.\n";

/* Long options */
static struct option long_options[] = {
	{NULL, 0, 0, 0}
};


/**********************************************************************
 * Logging functions
 */

void info(const char *fmt, ...)
{
	va_list ap;
    
    if (!verbose_flag)
        return;

	va_start(ap, fmt);
	vprintf(fmt, ap);
	printf("\n");
	fflush(stdout);
	va_end(ap);
}

void error(const char *fmt, ...)
{
	char *p, *msg;
	va_list ap;

    if (quiet_flag)
		return;

	va_start(ap, fmt);
	msg = g_strdup_vprintf(fmt, ap);
	va_end(ap);

	/* If the command name is the same as the name of the failed method,
	 * we may get duplication at the beginning of the message, like
	 * mv: mv: Destination entry: "/foo" already exists!
	 * Remove the duplication. */
	p = strchr(msg, ':');
	if (p)
	{
		int len = p - msg + 1;
		if (msg[len + 1] && !strncmp(msg, msg + len + 1, len))
			memmove(msg, msg + len + 1, strlen(msg) - len);
	}
	fprintf(stderr, "%s\n", msg);
	fflush(stderr);
	g_free(msg);
}

/**********************************************************************
 * Main routine for the gLite channel management tools
 */

int main(int argc, char *argv[])
{
    glite_delegation_ctx *ctx;
	char *options, *prog_name;
	int optlen, ret, c;
	int options_index = 0;

	/* Determine the program name we were called as */
	prog_name = strrchr(argv[0], '/');
	if (!prog_name)
		prog_name = argv[0];
	else
		prog_name++;

	/* Build the option string for getopt() */
	optlen = strlen(common_options) + 1;
	if (tool_options)
		optlen += strlen(tool_options);

	options = malloc(optlen);
	if (!options)
	{
		fprintf(stderr, "Out of memory\n");
		exit(-1);
	}

	strcpy(options, common_options);
	if (tool_options)
		strcat(options, tool_options);

	/* Parse the command line options */
	while ((c = getopt_long(argc, argv, options, long_options, &options_index)) != -1) switch (c)
	{
		case 'h':
			printf("Usage: %s [options]", prog_name);
			if (tool_usage)
				printf(" %s\n", tool_usage);
			else
				printf("\n");

			printf("Available options:\n%s", common_help);
			if (tool_help)
				printf("%s", tool_help);
			exit(0);
		case 's':
			service_location = optarg;
			break;
		case 'V':
			printf("%s\n", PACKAGE_VERSION);
			exit(0);
        case 'v':
            verbose_flag = 1;
            break;
        case 'q':
            quiet_flag = 1;
            break;
		case ':':
			fprintf(stderr, "Option argument is missing\n");
			exit(EXIT_FAILURE);
		case '?':
			fprintf(stderr, "Unknown command line option\n");
			exit(EXIT_FAILURE);
		default:
			ret = tool_parse_cmdline(c, optarg);
			if (ret)
			{
				fprintf(stderr, "Illegal command line arguments\n");
				exit(EXIT_FAILURE);
			}
			break;
	}

	/* Drop the processed options from the argument list */
	free(options);
	argc -= optind;
	argv += optind;

    ctx = glite_delegation_new(service_location);
    if(!ctx) 
    {
        fprintf(stderr, "Error: %s cannot initialize the connection", prog_name);
        exit(EXIT_FAILURE);
    }

    if (verbose_flag) {
        const char *endpoint = glite_delegation_get_endpoint(ctx);
        const char *version;
        if(endpoint)
            info("# Service endpoint: %s", endpoint);

        info("# Client version: %s", PACKAGE_VERSION);
        info("# Client interface version: %s", VERSION);
        
        version = glite_delegation_getVersion(ctx);
        if(version) {
            info("# Service version: %s", version);
            free(version);
        }

        version = glite_delegation_getInterfaceVersion(ctx);
        if (version) {
            info("# Service interface version: %s", version);
            free(version);
        }
    }

	/* Perform the operation */
	ret = tool_doit(ctx, argc, argv);

    glite_delegation_free(ctx);

	return ret;
}
